import sys


from formula import *
from formulawalker import *

from common import toBytes
from fractions import Fraction



def fixSMTv1(txt):
    import re

    res = re.sub(r"\(- (\d+(\.\d*)?)\)", r"(~ \1)", txt)

    return res



class SmtPrinter(Walker):

    def __init__(self, stream, strictSyntax=True, useBytes=False, smt2version=True):
        Walker.__init__(self)
        self.stream = stream
        self.strictSyntax = strictSyntax

        self.tb = lambda x: x
        if useBytes:
            self.tb = toBytes

        self.smt2version = smt2version
        return


    def printer(self, f):
        self.walk(f)
        return


    def walkAnd(self, formula):
        if len(formula.getSons()) > 1:
            self.stream.write(self.tb("(and"))
            for s in formula.getSons():
                self.stream.write(self.tb(" "))
                self.walk(s)
            self.stream.write(self.tb(")"))
        elif len(formula.getSons()) == 0:
            self.stream.write(self.tb("true"))
        else:
           for s in formula.getSons():
               self.walk(s)
        return


    def walkOr(self, formula):
        if len(formula.getSons()) > 1:
            self.stream.write(self.tb("(or"))
            for s in formula.getSons():
                self.stream.write(self.tb(" "))
                self.walk(s)
            self.stream.write(self.tb(")"))
        elif len(formula.getSons()) == 0:
            self.stream.write(self.tb("false"))
        else:
           for s in formula.getSons():
               self.walk(s)
        return


    def walkNot(self, formula):
        self.stream.write(self.tb("(not "))
        self.walk(formula.formula)
        self.stream.write(self.tb(")"))
        return


    def walkSymbol(self, formula):
        self.stream.write(self.tb(formula.name))
        return


    def walkPlus(self, formula):
        self.stream.write(self.tb("(+ "))
        self.walk(formula.left)
        self.stream.write(self.tb(" "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkTimes(self, formula):
        self.stream.write(self.tb("(* "))
        self.walk(formula.left)
        self.stream.write(self.tb(" "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return

    def walkDiv(self, formula):
        self.stream.write(self.tb("(/ "))
        self.walk(formula.left)
        self.stream.write(self.tb(" "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkMinus(self, formula):
        self.stream.write(self.tb("(- "))
        self.walk(formula.left)
        self.stream.write(self.tb(" "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkEquals(self, formula):
        self.stream.write(self.tb("(= "))
        self.walk(formula.left)
        self.stream.write(self.tb(" "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkGe(self, formula):
        self.stream.write(self.tb("(>= "))
        self.walk(formula.left)
        self.stream.write(self.tb(" "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkLe(self, formula):
        self.stream.write(self.tb("(<= "))
        self.walk(formula.left)
        self.stream.write(self.tb(" "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkGt(self, formula):
        self.stream.write(self.tb("(> "))
        self.walk(formula.left)
        self.stream.write(self.tb(" "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkLt(self, formula):
        self.stream.write(self.tb("(< "))
        self.walk(formula.left)
        self.stream.write(self.tb(" "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkFunction(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.name)
        for p in formula.params:
            self.stream.write(self.tb(" "))
            self.walk(p)
        self.stream.write(self.tb(")"))
        return


    def walkConst(self, formula):
        if type(formula.value) == tuple:
            (n,d) = formula.value
            res = str(n)
            if d != 1:
                res = "(/ " + res + " " + str(d) + ")"
            self.stream.write(self.tb(res))
            return
        else:
           if formula.value < 0:
               # Unary minus
               if self.smt2version:
                   self.stream.write(self.tb("(- " + str(-formula.value) + ")"))
               else:
                   self.stream.write(self.tb("(~ " + str(-formula.value) + ")"))
           else:
               self.stream.write(self.tb(str(formula.value)))
        return


    def walkBConst(self, formula):
        if formula.value:
            self.stream.write(self.tb("true"))
        else:
            self.stream.write(self.tb("false"))
        return


    def walkForall(self, formula):
        if len(formula.variables) > 0:
            self.stream.write(self.tb("(forall"))

            if self.strictSyntax:
                self.stream.write(self.tb(" ("))

            for s in formula.variables:
                self.stream.write(self.tb(" ("))
                self.walk(s)
                self.stream.write(self.tb(" %s)" % s.typename.toString(False)))

            if self.strictSyntax:
                self.stream.write(self.tb(") "))

            self.walk(formula.formula)

            self.stream.write(self.tb(")"))
        else:
            self.walk(formula.formula)
        return


    def walkExists(self, formula):
        if len(formula.variables) > 0:
            self.stream.write(self.tb("(exists"))

            if self.strictSyntax:
                self.stream.write(self.tb(" ("))

            for s in formula.variables:
                self.stream.write(self.tb(" ("))
                self.walk(s)
                self.stream.write(self.tb(" %s)" % s.typename.toString(False)))

            if self.strictSyntax:
                self.stream.write(self.tb(") "))

            self.walk(formula.formula)

            self.stream.write(self.tb(")"))
        else:
            self.walk(formula.formula)
        return


    def walkDefinition(self, formula):
        self.stream.write(self.tb("(let"))

        if self.strictSyntax:
            self.stream.write(self.tb(" ("))

        self.stream.write(self.tb(" (%s " % formula.name))
        self.walk(formula.definition)
        self.stream.write(self.tb(")"))

        if self.strictSyntax:
            self.stream.write(self.tb(") "))

        self.walk(formula.formula)
        self.stream.write(self.tb(")") )
        return


    def walkTextFormula(self, formula):
        txt = formula.text

        if not self.smt2version:
            txt = fixSMTv1(txt)

        self.stream.write(self.tb(txt))
        return

    pass



class HRPrinter(Walker):

    def __init__(self, stream, useBytes=False):
        Walker.__init__(self)
        self.stream = stream

        self.tb = lambda x: x
        if useBytes:
            self.tb = toBytes
        return


    def printer(self, f):
        self.walk(f)
        return


    def walkAnd(self, formula):
        if len(formula.getSons()) > 1:
            self.stream.write(self.tb("("))
            sons = formula.getSons()
            count = 0
            for s in sons:
                self.walk(s)
                count += 1
                if count != len(sons):
                    self.stream.write(self.tb(" & "))
            self.stream.write(self.tb(")"))
        elif len(formula.getSons()) == 0:
            self.stream.write(self.tb("True"))
        else:
           for s in formula.getSons():
               self.walk(s)
        return


    def walkOr(self, formula):
        if len(formula.getSons()) > 1:
            self.stream.write(self.tb("("))
            sons = formula.getSons()
            count = 0
            for s in sons:
                self.walk(s)
                count += 1
                if count != len(sons):
                    self.stream.write(self.tb(" | "))
            self.stream.write(self.tb(")"))
        elif len(formula.getSons()) == 0:
            self.stream.write(self.tb("False"))
        else:
           for s in formula.getSons():
               self.walk(s)
        return


    def walkNot(self, formula):
        self.stream.write(self.tb("(! "))
        self.walk(formula.formula)
        self.stream.write(self.tb(")"))
        return


    def walkSymbol(self, formula):
        self.stream.write(self.tb(formula.name))
        return


    def walkPlus(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.left)
        self.stream.write(self.tb(" + "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkTimes(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.left)
        self.stream.write(self.tb(" * "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return

    def walkDiv(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.left)
        self.stream.write(self.tb(" / "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkMinus(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.left)
        self.stream.write(self.tb(" - "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkEquals(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.left)
        self.stream.write(self.tb(" = "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkGe(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.left)
        self.stream.write(self.tb(" >= "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkLe(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.left)
        self.stream.write(self.tb(" <= "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkGt(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.left)
        self.stream.write(self.tb(" > "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkLt(self, formula):
        self.stream.write(self.tb("("))
        self.walk(formula.left)
        self.stream.write(self.tb(" < "))
        self.walk(formula.right)
        self.stream.write(self.tb(")"))
        return


    def walkFunction(self, formula):
        self.walk(formula.name)
        self.stream.write(self.tb("("))
        count = 0
        for p in formula.params:
            self.walk(p)
            count += 1
            if count != len(formula.params):
                self.stream.write(self.tb(" "))
        self.stream.write(self.tb(")"))
        return


    def walkConst(self, formula):
        assert type(formula.value) == Fraction, \
            "The type was " + str(type(formula.value))
        self.stream.write(self.tb(str(formula.value)))
        return


    def walkBConst(self, formula):
        if formula.value:
            self.stream.write(self.tb("True"))
        else:
            self.stream.write(self.tb("False"))
        return


    def walkForall(self, formula):
        if len(formula.variables) > 0:
            self.stream.write(self.tb("(forall "))

            count = 0
            for s in formula.variables:
                self.walk(s)
                count += 1
                if count != len(formula.variables):
                    self.stream.write(self.tb(", "))

            self.stream.write(self.tb(" . "))

            self.walk(formula.formula)

            self.stream.write(self.tb(")"))
        else:
            self.walk(formula.formula)
        return


    def walkExists(self, formula):
        if len(formula.variables) > 0:
            self.stream.write(self.tb("(exists "))

            count = 0
            for s in formula.variables:
                self.walk(s)
                count += 1
                if count != len(formula.variables):
                    self.stream.write(self.tb(", "))

            self.stream.write(self.tb(" . "))

            self.walk(formula.formula)

            self.stream.write(self.tb(")"))
        else:
            self.walk(formula.formula)
        return


    def walkDefinition(self, formula):
        self.stream.write(self.tb("(let "))
        self.stream.write(self.tb("(%s = " % formula.name))
        self.walk(formula.definition)
        self.stream.write(self.tb(") "))

        self.walk(formula.formula)
        self.stream.write(self.tb(")") )
        return


    def walkTextFormula(self, formula):
        txt = formula.text
        self.stream.write(self.tb(txt))
        return

    pass


def printFormula(formula, stream=sys.stdout):
    print("---------")
    p = HRPrinter(stream, useBytes=(stream!=sys.stdout))
    p.printer(formula)
    print("\n---------")
    return
