import sys

from mathsat import *
from formula import *

import common
from encoders.printers import SmtPrinter

from fractions import Fraction

class MSatConverter(object):

    def __init__(self, env):
        self.env = env

        self.conversions = {}
        self.definitions = {}
        self.definitionsFormulae = []

        self.realType = msat_get_rational_type(env)
        return

    def safeState(self):
        return len(self.definitions) == 0

    def convert(self, formula):
        res = None
        env = self.env

        sons = list(formula.getSons())

        # Memoization termination case
        if formula in self.conversions and len(self.definitions) == 0:
            return self.conversions[formula]


        if formula.nodeType == FNode.AND:
            if len(sons) == 0:
                res = msat_make_true(env)
            else:
                res = self.convert(sons[0])
                for i in range(1, len(sons)):
                    res = msat_make_and(env, res, self.convert(sons[i]))

        elif formula.nodeType == FNode.OR:
            if len(sons) == 0:
                res = msat_make_false(env)
            else:
                res = self.convert(sons[0])
                for i in range(1, len(sons)):
                    res = msat_make_or(env, res, self.convert(sons[i]))

        elif formula.nodeType == FNode.NOT:
            assert len(sons) == 1
            res = msat_make_not(env, self.convert(sons[0]))

        elif formula.nodeType == FNode.SYMBOL:
            if formula.name in self.definitions:
                res = self.definitions[formula.name]
            else:
                res = msat_make_constant(env, msat_find_decl(env, formula.name))

        elif formula.nodeType == FNode.GE:
            rew = LE(Times(Constant(-1), formula.left), Times(Constant(-1), formula.right))
            res = self.convert(rew)

        elif formula.nodeType == FNode.LE:
            res = msat_make_leq(env, self.convert(sons[0]), self.convert(sons[1]))

        elif formula.nodeType == FNode.GT:
            rew = Not(LE(formula.left, formula.right))
            res = self.convert(rew)

        elif formula.nodeType == FNode.LT:
            rew = And([Not(Equals(formula.left, formula.right)), LE(formula.left, formula.right)])
            res = self.convert(rew)

        elif formula.nodeType == FNode.CONSTANT:
            assert type(formula.value) == Fraction
            frac = formula.value
            n,d = frac.numerator, frac.denominator
            res = msat_make_number(env, str(n) + "/" + str(d))

        elif formula.nodeType == FNode.BCONSTANT:
            if formula.value:
                res = msat_make_true(env)
            else:
                res = msat_make_false(env)

        elif formula.nodeType == FNode.FORALL:
            assert False

        elif formula.nodeType == FNode.EXISTS:
            assert False

        elif formula.nodeType == FNode.DEFINITION:
            definition = self.convert(formula.definition)
            self.definitions[formula.name] = definition
            self.definitionsFormulae.append((formula.name, formula))
            res = self.convert(formula.formula)
            del self.definitions[formula.name]
            self.definitionsFormulae.pop()

        elif formula.nodeType == FNode.PLUS:
           res = msat_make_plus(env, self.convert(sons[0]), self.convert(sons[1]))

        elif formula.nodeType == FNode.MINUS:
           res = msat_make_plus(env,
                                self.convert(sons[0]),
                                msat_make_times(env, msat_make_number(env, "-1"), self.convert(sons[1])))

        elif formula.nodeType == FNode.EQUALS:
            res = msat_make_equal(env,
                                  self.convert(sons[0]),
                                  self.convert(sons[1]))

        elif formula.nodeType == FNode.TIMES:
            res = msat_make_times(env, self.convert(sons[0]), self.convert(sons[1]))

        elif formula.nodeType == FNode.DIV:
            if formula.right.nodeType == FNode.CONSTANT:
                val = formula.right.value
                oth = formula.left
            else:
                assert formula.left.nodeType == FNode.CONSTANT
                val = formula.left.value
                oth = formula.right

            assert type(val) == Fraction

            mul = msat_make_number(env, str(val.denominator) + "/" + str(val.numerator))
            res = msat_make_times(env, mul, self.convert(oth))


        elif formula.nodeType == FNode.FUNCTION:
            assert False

        elif formula.nodeType == FNode.TEXTFORMULA:
            string = str(formula.text)

            for idx in range(len(self.definitionsFormulae)-1, -1, -1):
                (d,t) = self.definitionsFormulae[idx]
                buf = common.getIOBuffer()
                p = SmtPrinter(buf, strictSyntax=True, useBytes=True, smt2version=True)
                p.printer(t.definition)
                strformula = buf.getvalue()
                buf.close()
                strformula = common.convertString(strformula)
                string = string.replace(d, strformula)

            res = msat_from_string(env, string.encode('utf-8'))


        if res is None or MSAT_ERROR_TERM(res):
            print("Error in conversion to MathSAT!\n")
            print("Unable to convert the following formula:")
            p = SmtPrinter(sys.stdout, strictSyntax=True, useBytes=False, smt2version=True)
            p.printer(formula)
            print("\nThe reported error is: " + msat_last_error_message(env))
            exit(1)

        self.conversions[formula] = res
        return res

    pass
