#!/usr/bin/env python

import argparse
import os.path
import sys

sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), "lib"))
sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), "bin", "pylib"))
sys.setrecursionlimit(100000)

from structures import Problem, AtomicConstraint
from typing import Types
from formula import Environment
from extractors.solvers.msat import *


class PVYS(object):

    def __init__(self, environment):
        self.environment = environment
        self.solver = None
        return


    def solve(self, problem, model=False):
        st = self.environment.symbolTable

        Cs = set()
        Cc = set()
        Ce = set()

        for c in problem.constraints:
            if c.isSimple() or not c.hasUncontrollables():
                # all simple constraints + constraints on controllable variables
                Cs.add(c)
            elif c.contingent:
                # Disjunctive contingent constraints
                Cc.add(c)
            else:
                # Disjunctive free constraints with at least one uncontrollable
                Ce.add(c)


        variables = [x for x in problem.timePoints.values() if x.controllable]
        fvars = []
        self.solver = MathSAT5Solver(self.environment)
        for tp in variables:
            v = st.createNewSymbol(tp.name, Types.REAL)
            self.solver.declareVariable(v)
            fvars.append(v)

        res = self.dtpu_sc(BConstant(True), set(), Cs, Cc, Ce)

        val = self.is_consistent(res, model=model, variables=fvars)

        self.solver.destroy()
        self.solver = None

        if model:
            if val[0]:
                print "--- Strong Schedule ---"
                for k, v in val[1].items():
                    print "Schedule %s at %s" % (k.name, v)
                print "-----------------------"
                print ""
            return val[0]
        else:
            return val

    def select_variable(self, s):
        assert len(s) > 0
        for v in s:
            return v


    def disjuncts(self, c):
        return c.atoms


    def sc_transform(self, prob, c):
        def get_contingency(t):
            for a in prob:
                if a.contingent and a.dest == t:
                    return (a.source, a.lowerBound, a.upperBound)
            return (None, None, None)

        def plus(a, b):
            assert b != common.INFINITY and b != common.MINUS_INFINITY
            if a == common.INFINITY or a == common.MINUS_INFINITY:
                return a

            return a + b


        toprocess = []
        if c.contingent:
            for a in prob:
                if a != c and (a.source == c.dest or a.dest == c.dest):
                    toprocess.append(a)
        else:
            toprocess.append(c)

        result = set()
        for p in toprocess:
            if p.source.controllable and p.dest.controllable:
                result.add(c)
            elif p.dest.controllable:
                # Free of type 2
                (s, l, u) = get_contingency(p.source)
                if s is not None:
                    res = AtomicConstraint(s, p.dest,
                                           plus(p.lowerBound, u),
                                           plus(p.upperBound, l),
                                           False)
                    result.add(res)

            elif p.source.controllable:
                # Free of type 3
                (s, l, u) = get_contingency(p.dest)
                if s is not None:
                    res = AtomicConstraint(p.source, s,
                                           plus(p.lowerBound, -l),
                                           plus(p.upperBound, -u),
                                           False)
                    result.add(res)

            else:
                # Free of type 4
                (s1, l1, u1) = get_contingency(p.dest)
                (s2, l2, u2) = get_contingency(p.source)
                if s1 is not None and s2 is not None:
                    res = AtomicConstraint(s1, s2,
                                           plus(p.lowerBound, u2 - l1),
                                           plus(p.upperBound, l2 - u1),
                                           False)
                    result.add(res)
        return result


    def minimal_network(self, prob):
        # No need for minimal network in SMT
        return prob


    def is_consistent(self, prob, model=False, variables=None):
        assert self.solver is not None
        st = self.environment.symbolTable

        self.solver.push()
        self.solver.addAssertion(prob)
        res = self.solver.solve()
        m = None

        if res and model:
            assert variables is not None
            m = {}
            for v in variables:
                val = self.solver.getModelValue(v)
                m[v] = val
        self.solver.pop()

        if model:
            return res, m
        else:
            return res


    def to_formulae(self, s):
        st = self.environment.symbolTable
        res = []

        for a in s:
            vs = st.getSymbolByName(a.source.name)
            vd = st.getSymbolByName(a.dest.name)
            l = a.lowerBound
            u = a.upperBound

            if l != common.MINUS_INFINITY:
                res.append(GE(Minus(vd,vs), Constant(l)))
            if u != common.INFINITY:
                res.append(LE(Minus(vd,vs), Constant(u)))
        return res


    def dtpu_sc(self, A, Ac, Cs, Cc, Ce):
        if len(Cs) == 0:
            G = self.minimal_network(A)
            return self.all_paths_sc(A, Ac, Cc, Ce, G)
        else:
            Ci = self.select_variable(Cs)

            Cs1 = Cs - set([Ci])

            for cij in self.disjuncts(Ci):
                Ac1 = Ac | set([cij])

                # Conjoin new constraints in STP A
                A1 = And([A, And(self.to_formulae(self.sc_transform(Ac1, cij)))])

                if self.is_consistent(A1):
                    S = self.dtpu_sc(A1, Ac1, Cs1, Cc, Ce)
                    if self.is_consistent(S):
                        return S
        return BConstant(False)


    def all_paths_sc(self, A, Ac, Cc, Ce, G):
        if len(Cc) == 0:
            G = And([G, self.satisfy_ce(A, Ac, Ce)])
        else:
            Ci = self.select_variable(Cc)
            Cc1 = Cc - set([Ci])
            for cij in self.disjuncts(Ci):
                Ac1 = Ac | set([cij])
                A1 = And([A, And(self.to_formulae(self.sc_transform(Ac1, cij)))])
                if self.is_consistent(A1):
                    G = self.all_paths_sc(A1, Ac1, Cc1, Ce, G)
                    if not self.is_consistent(G):
                        return BConstant(False)
                else:
                    return BConstant(False)
        return G


    def satisfy_ce(self, A, Ac, Ce):
        H = BConstant(False)
        if len(Ce) == 0:
            H = self.minimal_network(A)
        else:
            Ci = self.select_variable(Ce)
            Ce1 = Ce - set([Ci])
            for cij in self.disjuncts(Ci):
                A1 = And([A, And(self.to_formulae(self.sc_transform(Ac, cij)))])
                printFormula(A1)
                if self.is_consistent(A1):
                    H = Or([H, self.satisfy_ce(A1, Ac, Ce1)])
        return H

#End of PVYS




# Main script
if __name__ == "__main__":

    parser = \
        argparse.ArgumentParser(description='PVYS Strong Controllability Solver')

    parser.add_argument('inputFile', metavar='inputFile', type=str,
                        help='The probmem to solve')
    parser.add_argument('--model', '-m', action='store_true',
                        help='Print a strong controllability strategy')

    args = parser.parse_args()




    env = Environment()

    tn = Problem()
    tn.fromFile(args.inputFile)


    s = PVYS(env)
    res = s.solve(tn, model=args.model)
    if res:
        print "sat"
    else:
        print "unsat"
